#'
#' Principal Components Analysis
#'
#' Principal components analysis of a H2O dataset using the power method
#' to calculate the singular value decomposition of the Gram matrix.
#'
#'
#' @param training_frame An \linkS4class{H2OFrame} object containing the
#'        variables in the model.
#' @param x (Optional) A vector containing the data columns on which SVD operates.
#' @param k The number of principal components to be computed. This must be
#'        between 1 and min(ncol(training_frame), nrow(training_frame)) inclusive.
#' @param model_id (Optional) The unique hex key assigned to the
#'        resulting model. Automatically generated if none is provided.
#' @param max_iterations The maximum number of iterations to run each power
#'        iteration loop. Must be between 1 and 1e6 inclusive.
#' @param transform A character string that indicates how the training data
#'        should be transformed before running PCA. Possible values are "NONE":
#'        for no transformation, "DEMEAN": for subtracting the mean of each
#'        column, "DESCALE": for dividing by the standard deviation of each
#'        column, "STANDARDIZE": for demeaning and descaling, and "NORMALIZE":
#'        for demeaning and dividing each column by its range (max - min).
#' @param pca_method A character string that indicates how PCA should be calculated.
#'        Possible values are "GramSVD": distributed computation of the Gram matrix
#'        followed by a local SVD using the JAMA package, "Power": computation of
#'        the SVD using the power iteration method, "GLRM": fit a generalized low
#'        rank model with an l2 loss function (no regularization) and solve for the
#'        SVD using local matrix algebra.
#' @param seed (Optional) Random seed used to initialize the right singular vectors
#'        at the beginning of each power method iteration.
#' @param use_all_factor_levels (Optional) A logical value indicating whether all
#'        factor levels should be included in each categorical column expansion.
#'        If FALSE, the indicator column corresponding to the first factor level
#'        of every categorical variable will be dropped. Defaults to FALSE.
#' @return Returns an object of class \linkS4class{H2ODimReductionModel}.
#' @seealso \code{\link{h2o.svd}}, \code{\link{h2o.glrm}}
#' @examples
#' library(h2o)
#' localH2O <- h2o.init()
#' ausPath <- system.file("extdata", "australia.csv", package="h2o")
#' australia.hex <- h2o.uploadFile(localH2O, path = ausPath)
#' h2o.prcomp(training_frame = australia.hex, k = 8, transform = "STANDARDIZE")
#' @export
h2o.prcomp <- function(training_frame, x, k,
                      ## AUTOGENERATED PARAMETERS ##     # these defaults are not read by h2o
                      model_id,                          # h2o generates its own default parameters
                      max_iterations = 1000,
                      transform = c("NONE", "DEMEAN", "DESCALE", "STANDARDIZE"),
                      pca_method = c("GramSVD", "Power", "GLRM"),
                      seed,
                      use_all_factor_levels)
{
  # Required args: training_frame
  if( missing(training_frame) ) stop("argument \"training_frame\" is missing, with no default")

  # Training_frame may be a key or an H2OFrame object
  if (!inherits(training_frame, "H2OFrame"))
    tryCatch(training_frame <- h2o.getFrame(training_frame),
             error = function(err) {
               stop("argument \"training_frame\" must be a valid H2OFrame or key")
             })

  ## -- Force evaluate temporary ASTs -- ##
  delete <- !.is.eval(training_frame)
  if( delete ) {
    temp_key <- training_frame@frame_id
    .h2o.eval.frame(conn = training_frame@conn, ast = training_frame@mutable$ast, frame_id = temp_key)
  }

  # Gather user input
  parms <- list()
  parms$training_frame <- training_frame
  if(!missing(x))
    parms$ignored_columns <- .verify_datacols(training_frame, x)$cols_ignore
  if(!missing(k))
    parms$k <- as.numeric(k)    # TODO: Want default to min(n,p) where n = nrow, p = ncol of adapted training_frame
  if(!missing(model_id))
    parms$model_id <- model_id
  if(!missing(max_iterations))
    parms$max_iterations <- max_iterations
  if(!missing(transform))
    parms$transform <- transform
  if(!missing(pca_method))
    parms$pca_method <- pca_method
  if(!missing(seed))
    parms$seed <- seed
  if(!missing(use_all_factor_levels))
    parms$use_all_factor_levels <- use_all_factor_levels

  # Error check and build model
  .h2o.createModel(training_frame@conn, 'pca', parms, 99)
}
